from __future__ import annotations

import contextlib
import logging
from asyncio import CancelledError
from contextlib import contextmanager
from threading import Thread
from typing import TYPE_CHECKING, Any, Callable, cast

import psycopg
import psycopg.errors
import psycopg_pool
from psycopg import Connection, Cursor, Error
from psycopg.generators import notifies
from psycopg.rows import DictRow, dict_row
from psycopg.sql import SQL, Composed, Identifier
from typing_extensions import TypeVar

from eventsourcing.persistence import (
    AggregateRecorder,
    ApplicationRecorder,
    DatabaseError,
    DataError,
    InfrastructureFactory,
    IntegrityError,
    InterfaceError,
    InternalError,
    ListenNotifySubscription,
    Notification,
    NotSupportedError,
    OperationalError,
    PersistenceError,
    ProcessRecorder,
    ProgrammingError,
    StoredEvent,
    Subscription,
    Tracking,
    TrackingRecorder,
)
from eventsourcing.utils import Environment, EnvType, resolve_topic, retry, strtobool

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from uuid import UUID

    from psycopg.abc import Query
    from typing_extensions import Self

logging.getLogger("psycopg.pool").setLevel(logging.ERROR)
logging.getLogger("psycopg").setLevel(logging.ERROR)

# Copy of "private" psycopg.errors._NO_TRACEBACK (in case it changes)
# From psycopg: "Don't show a complete traceback upon raising these exception.
# Usually the traceback starts from internal functions (for instance in the
# server communication callbacks) but, for the end user, it's more important
# to get the high level information about where the exception was raised, for
# instance in a certain `Cursor.execute()`."
NO_TRACEBACK = (Error, KeyboardInterrupt, CancelledError)


class ConnectionPool(psycopg_pool.ConnectionPool[Any]):
    def __init__(
        self,
        *args: Any,
        get_password_func: Callable[[], str] | None = None,
        **kwargs: Any,
    ) -> None:
        self.get_password_func = get_password_func
        super().__init__(*args, **kwargs)

    def _connect(self, timeout: float | None = None) -> Connection[Any]:
        if self.get_password_func:
            self.kwargs["password"] = self.get_password_func()
        return super()._connect(timeout=timeout)


class PostgresDatastore:
    def __init__(  # noqa: PLR0913
        self,
        dbname: str,
        host: str,
        port: str | int,
        user: str,
        password: str,
        *,
        connect_timeout: float = 5.0,
        idle_in_transaction_session_timeout: float = 0,
        pool_size: int = 1,
        max_overflow: int = 0,
        max_waiting: int = 0,
        conn_max_age: float = 60 * 60.0,
        pre_ping: bool = False,
        lock_timeout: int = 0,
        schema: str = "",
        pool_open_timeout: float | None = None,
        get_password_func: Callable[[], str] | None = None,
        single_row_tracking: bool = True,
    ):
        self.idle_in_transaction_session_timeout = idle_in_transaction_session_timeout
        self.pre_ping = pre_ping
        self.pool_open_timeout = pool_open_timeout
        self.single_row_tracking = single_row_tracking

        check = ConnectionPool.check_connection if pre_ping else None
        self.pool = ConnectionPool(
            get_password_func=get_password_func,
            connection_class=Connection[DictRow],
            kwargs={
                "dbname": dbname,
                "host": host,
                "port": port,
                "user": user,
                "password": password,
                "row_factory": dict_row,
            },
            min_size=pool_size,
            max_size=pool_size + max_overflow,
            open=False,
            configure=self.after_connect_func(),
            timeout=connect_timeout,
            max_waiting=max_waiting,
            max_lifetime=conn_max_age,
            check=check,
        )
        self.lock_timeout = lock_timeout
        self.schema = schema.strip() or "public"

    def after_connect_func(self) -> Callable[[Connection[Any]], None]:
        statement = SQL("SET idle_in_transaction_session_timeout = '{0}ms'").format(
            int(self.idle_in_transaction_session_timeout * 1000)
        )

        def after_connect(conn: Connection[DictRow]) -> None:
            conn.autocommit = True
            conn.cursor().execute(statement)

        return after_connect

    @contextmanager
    def get_connection(self) -> Iterator[Connection[DictRow]]:
        try:
            wait = self.pool_open_timeout is not None
            timeout = self.pool_open_timeout or 30.0
            self.pool.open(wait, timeout)

            with self.pool.connection() as conn:
                yield conn
        except psycopg.InterfaceError as e:
            # conn.close()
            raise InterfaceError(str(e)) from e
        except psycopg.OperationalError as e:
            # conn.close()
            raise OperationalError(str(e)) from e
        except psycopg.DataError as e:
            raise DataError(str(e)) from e
        except psycopg.IntegrityError as e:
            raise IntegrityError(str(e)) from e
        except psycopg.InternalError as e:
            raise InternalError(str(e)) from e
        except psycopg.ProgrammingError as e:
            raise ProgrammingError(str(e)) from e
        except psycopg.NotSupportedError as e:
            raise NotSupportedError(str(e)) from e
        except psycopg.DatabaseError as e:
            raise DatabaseError(str(e)) from e
        except psycopg.Error as e:
            # conn.close()
            raise PersistenceError(str(e)) from e
        except Exception:
            # conn.close()
            raise

    @contextmanager
    def transaction(self, *, commit: bool = False) -> Iterator[Cursor[DictRow]]:
        with self.get_connection() as conn, conn.transaction(force_rollback=not commit):
            yield conn.cursor()

    def close(self) -> None:
        self.pool.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()


class PostgresRecorder:
    """Base class for recorders that use PostgreSQL."""

    MAX_IDENTIFIER_LEN = 63
    # From the PostgreSQL docs: "The system uses no more than NAMEDATALEN-1 bytes
    # of an identifier; longer names can be written in commands, but they will be
    # truncated. By default, NAMEDATALEN is 64 so the maximum identifier length is
    # 63 bytes." https://www.postgresql.org/docs/current/sql-syntax-lexical.html

    def __init__(
        self,
        datastore: PostgresDatastore,
    ):
        self.datastore = datastore
        self.create_table_statements = self.construct_create_table_statements()

    @staticmethod
    def check_table_name_length(table_name: str) -> None:
        if len(table_name) > PostgresRecorder.MAX_IDENTIFIER_LEN:
            msg = f"Table name too long: {table_name}"
            raise ProgrammingError(msg)

    def construct_create_table_statements(self) -> list[Composed]:
        return []

    def create_table(self) -> None:
        with self.datastore.transaction(commit=True) as curs:
            self._create_table(curs)

    def _create_table(self, curs: Cursor[DictRow]) -> None:
        for statement in self.create_table_statements:
            curs.execute(statement, prepare=False)


class PostgresAggregateRecorder(PostgresRecorder, AggregateRecorder):
    def __init__(
        self,
        datastore: PostgresDatastore,
        *,
        events_table_name: str = "stored_events",
    ):
        super().__init__(datastore)
        self.check_table_name_length(events_table_name)
        self.events_table_name = events_table_name
        # Index names can't be qualified names, but
        # are created in the same schema as the table.
        self.notification_id_index_name = (
            f"{self.events_table_name}_notification_id_idx"
        )
        self.create_table_statements.append(
            SQL(
                "CREATE TABLE IF NOT EXISTS {0}.{1} ("
                "originator_id uuid NOT NULL, "
                "originator_version bigint NOT NULL, "
                "topic text, "
                "state bytea, "
                "PRIMARY KEY "
                "(originator_id, originator_version)) "
                "WITH (autovacuum_enabled=false)"
            ).format(
                Identifier(self.datastore.schema),
                Identifier(self.events_table_name),
            )
        )

        self.insert_events_statement = SQL(
            "INSERT INTO {0}.{1} VALUES (%s, %s, %s, %s)"
        ).format(
            Identifier(self.datastore.schema),
            Identifier(self.events_table_name),
        )

        self.select_events_statement = SQL(
            "SELECT * FROM {0}.{1} WHERE originator_id = %s"
        ).format(
            Identifier(self.datastore.schema),
            Identifier(self.events_table_name),
        )

        self.lock_table_statements: list[Query] = []

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def insert_events(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        exc: Exception | None = None
        notification_ids: Sequence[int] | None = None
        with self.datastore.get_connection() as conn:
            with conn.pipeline() as pipeline, conn.transaction():
                # Do other things first, so they can be pipelined too.
                with conn.cursor() as curs:
                    self._insert_events(curs, stored_events, **kwargs)
                # Then use a different cursor for the executemany() call.
                with conn.cursor() as curs:
                    try:
                        self._insert_stored_events(curs, stored_events, **kwargs)
                        # Sync now, so any uniqueness constraint violation causes an
                        # IntegrityError to be raised here, rather an InternalError
                        # being raised sometime later e.g. when commit() is called.
                        pipeline.sync()
                        notification_ids = self._fetch_ids_after_insert_events(
                            curs, stored_events, **kwargs
                        )
                    except Exception as e:
                        # Avoid psycopg emitting a pipeline warning.
                        exc = e
            if exc:
                # Reraise exception after pipeline context manager has exited.
                raise exc
        return notification_ids

    def _insert_events(
        self,
        curs: Cursor[DictRow],
        stored_events: Sequence[StoredEvent],
        **_: Any,
    ) -> None:
        pass

    def _insert_stored_events(
        self,
        curs: Cursor[DictRow],
        stored_events: Sequence[StoredEvent],
        **_: Any,
    ) -> None:
        # Only do something if there is something to do.
        if len(stored_events) > 0:
            self._lock_table(curs)

            self._notify_channel(curs)

            # Insert events.
            curs.executemany(
                query=self.insert_events_statement,
                params_seq=[
                    (
                        stored_event.originator_id,
                        stored_event.originator_version,
                        stored_event.topic,
                        stored_event.state,
                    )
                    for stored_event in stored_events
                ],
                returning="RETURNING" in self.insert_events_statement.as_string(),
            )

    def _lock_table(self, curs: Cursor[DictRow]) -> None:
        pass

    def _notify_channel(self, curs: Cursor[DictRow]) -> None:
        pass

    def _fetch_ids_after_insert_events(
        self,
        curs: Cursor[DictRow],
        stored_events: Sequence[StoredEvent],
        **kwargs: Any,
    ) -> Sequence[int] | None:
        return None

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def select_events(
        self,
        originator_id: UUID | str,
        *,
        gt: int | None = None,
        lte: int | None = None,
        desc: bool = False,
        limit: int | None = None,
    ) -> Sequence[StoredEvent]:
        statement = self.select_events_statement
        params: list[Any] = [originator_id]
        if gt is not None:
            params.append(gt)
            statement += SQL(" AND originator_version > %s")
        if lte is not None:
            params.append(lte)
            statement += SQL(" AND originator_version <= %s")
        statement += SQL(" ORDER BY originator_version")
        if desc is False:
            statement += SQL(" ASC")
        else:
            statement += SQL(" DESC")
        if limit is not None:
            params.append(limit)
            statement += SQL(" LIMIT %s")

        with self.datastore.get_connection() as conn, conn.cursor() as curs:
            curs.execute(statement, params, prepare=True)
            return [
                StoredEvent(
                    originator_id=row["originator_id"],
                    originator_version=row["originator_version"],
                    topic=row["topic"],
                    state=bytes(row["state"]),
                )
                for row in curs.fetchall()
            ]


class PostgresApplicationRecorder(PostgresAggregateRecorder, ApplicationRecorder):
    def __init__(
        self,
        datastore: PostgresDatastore,
        *,
        events_table_name: str = "stored_events",
    ):
        super().__init__(datastore, events_table_name=events_table_name)
        self.create_table_statements[-1] = SQL(
            "CREATE TABLE IF NOT EXISTS {0}.{1} ("
            "originator_id uuid NOT NULL, "
            "originator_version bigint NOT NULL, "
            "topic text, "
            "state bytea, "
            "notification_id bigserial, "
            "PRIMARY KEY "
            "(originator_id, originator_version)) "
            "WITH (autovacuum_enabled=false)"
        ).format(
            Identifier(self.datastore.schema),
            Identifier(self.events_table_name),
        )

        self.create_table_statements.append(
            SQL(
                "CREATE UNIQUE INDEX IF NOT EXISTS {0} "
                "ON {1}.{2} (notification_id ASC);"
            ).format(
                Identifier(self.notification_id_index_name),
                Identifier(self.datastore.schema),
                Identifier(self.events_table_name),
            )
        )

        self.channel_name = self.events_table_name.replace(".", "_")
        self.insert_events_statement = self.insert_events_statement + SQL(
            " RETURNING notification_id"
        )

        self.max_notification_id_statement = SQL(
            "SELECT MAX(notification_id) FROM {0}.{1}"
        ).format(
            Identifier(self.datastore.schema),
            Identifier(self.events_table_name),
        )

        self.lock_table_statements = [
            SQL("SET LOCAL lock_timeout = '{0}s'").format(self.datastore.lock_timeout),
            SQL("LOCK TABLE {0}.{1} IN EXCLUSIVE MODE").format(
                Identifier(self.datastore.schema),
                Identifier(self.events_table_name),
            ),
        ]

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def select_notifications(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> Sequence[Notification]:
        """Returns a list of event notifications
        from 'start', limited by 'limit'.
        """
        params: list[int | str | Sequence[str]] = []
        statement = SQL("SELECT * FROM {0}.{1}").format(
            Identifier(self.datastore.schema),
            Identifier(self.events_table_name),
        )
        has_where = False
        if start is not None:
            statement += SQL(" WHERE")
            has_where = True
            params.append(start)
            if inclusive_of_start:
                statement += SQL(" notification_id>=%s")
            else:
                statement += SQL(" notification_id>%s")

        if stop is not None:
            if not has_where:
                has_where = True
                statement += SQL(" WHERE")
            else:
                statement += SQL(" AND")

            params.append(stop)
            statement += SQL(" notification_id <= %s")

        if topics:
            # Check sequence and ensure list of strings.
            assert isinstance(topics, (tuple, list)), topics
            topics = list(topics) if isinstance(topics, tuple) else topics
            assert all(isinstance(t, str) for t in topics), topics
            if not has_where:
                statement += SQL(" WHERE")
            else:
                statement += SQL(" AND")
            params.append(topics)
            statement += SQL(" topic = ANY(%s)")

        params.append(limit)
        statement += SQL(" ORDER BY notification_id LIMIT %s")

        connection = self.datastore.get_connection()
        with connection as conn, conn.cursor() as curs:
            curs.execute(statement, params, prepare=True)
            return [
                Notification(
                    id=row["notification_id"],
                    originator_id=row["originator_id"],
                    originator_version=row["originator_version"],
                    topic=row["topic"],
                    state=bytes(row["state"]),
                )
                for row in curs.fetchall()
            ]

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def max_notification_id(self) -> int | None:
        """Returns the maximum notification ID."""
        with self.datastore.get_connection() as conn, conn.cursor() as curs:
            curs.execute(self.max_notification_id_statement)
            fetchone = curs.fetchone()
            assert fetchone is not None
            return fetchone["max"]

    def _lock_table(self, curs: Cursor[DictRow]) -> None:
        # Acquire "EXCLUSIVE" table lock, to serialize transactions that insert
        # stored events, so that readers don't pass over gaps that are filled in
        # later. We want each transaction that will be issued with notifications
        # IDs by the notification ID sequence to receive all its notification IDs
        # and then commit, before another transaction is issued with any notification
        # IDs. In other words, we want the insert order to be the same as the commit
        # order. We can accomplish this by locking the table for writes. The
        # EXCLUSIVE lock mode does not block SELECT statements, which acquire an
        # ACCESS SHARE lock, so the stored events table can be read concurrently
        # with writes and other reads. However, INSERT statements normally just
        # acquires ROW EXCLUSIVE locks, which risks the interleaving (within the
        # recorded sequence of notification IDs) of stored events from one transaction
        # with those of another transaction. And since one transaction will always
        # commit before another, the possibility arises when using ROW EXCLUSIVE locks
        # for readers that are tailing a notification log to miss items inserted later
        # but issued with lower notification IDs.
        # https://www.postgresql.org/docs/current/explicit-locking.html#LOCKING-TABLES
        # https://www.postgresql.org/docs/9.1/sql-lock.html
        # https://stackoverflow.com/questions/45866187/guarantee-monotonicity-of
        # -postgresql-serial-column-values-by-commit-order
        for lock_statement in self.lock_table_statements:
            curs.execute(lock_statement, prepare=True)

    def _notify_channel(self, curs: Cursor[DictRow]) -> None:
        curs.execute(SQL("NOTIFY {0}").format(Identifier(self.channel_name)))

    def _fetch_ids_after_insert_events(
        self,
        curs: Cursor[DictRow],
        stored_events: Sequence[StoredEvent],
        **kwargs: Any,
    ) -> Sequence[int] | None:
        notification_ids: list[int] = []
        len_events = len(stored_events)
        if len_events:
            while curs.nextset() and len(notification_ids) != len_events:
                if curs.statusmessage and curs.statusmessage.startswith("INSERT"):
                    row = curs.fetchone()
                    assert row is not None
                    notification_ids.append(row["notification_id"])
            if len(notification_ids) != len(stored_events):
                msg = "Couldn't get all notification IDs "
                msg += f"(got {len(notification_ids)}, expected {len(stored_events)})"
                raise ProgrammingError(msg)
        return notification_ids

    def subscribe(
        self, gt: int | None = None, topics: Sequence[str] = ()
    ) -> Subscription[ApplicationRecorder]:
        return PostgresSubscription(recorder=self, gt=gt, topics=topics)


class PostgresSubscription(ListenNotifySubscription[PostgresApplicationRecorder]):
    def __init__(
        self,
        recorder: PostgresApplicationRecorder,
        gt: int | None = None,
        topics: Sequence[str] = (),
    ) -> None:
        assert isinstance(recorder, PostgresApplicationRecorder)
        super().__init__(recorder=recorder, gt=gt, topics=topics)
        self._listen_thread = Thread(target=self._listen)
        self._listen_thread.start()

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        super().__exit__(*args, **kwargs)
        self._listen_thread.join()

    def _listen(self) -> None:
        try:
            with self._recorder.datastore.get_connection() as conn:
                conn.execute(
                    SQL("LISTEN {0}").format(Identifier(self._recorder.channel_name))
                )
                while not self._has_been_stopped and not self._thread_error:
                    # This block simplifies psycopg's conn.notifies(), because
                    # we aren't interested in the actual notify messages, and
                    # also we want to stop consuming notify messages when the
                    # subscription has an error or is otherwise stopped.
                    with conn.lock:
                        try:
                            if conn.wait(notifies(conn.pgconn), interval=0.1):
                                self._has_been_notified.set()
                        except NO_TRACEBACK as ex:  # pragma: no cover
                            raise ex.with_traceback(None) from None

        except BaseException as e:
            if self._thread_error is None:
                self._thread_error = e
            self.stop()


class PostgresTrackingRecorder(PostgresRecorder, TrackingRecorder):
    def __init__(
        self,
        datastore: PostgresDatastore,
        *,
        tracking_table_name: str = "notification_tracking",
        **kwargs: Any,
    ):
        super().__init__(datastore, **kwargs)
        self.check_table_name_length(tracking_table_name)
        self.tracking_table_name = tracking_table_name
        self.tracking_table_exists: bool = False
        self.tracking_migration_previous: int | None = None
        self.tracking_migration_current: int | None = None
        self.table_migration_identifier = "__migration__"
        self.has_checked_for_multi_row_tracking_table: bool = False
        if self.datastore.single_row_tracking:
            # For single-row tracking.
            self.create_table_statements.append(
                SQL(
                    "CREATE TABLE IF NOT EXISTS {0}.{1} ("
                    "application_name text, "
                    "notification_id bigint, "
                    "PRIMARY KEY "
                    "(application_name))"
                ).format(
                    Identifier(self.datastore.schema),
                    Identifier(self.tracking_table_name),
                )
            )
            self.insert_tracking_statement = SQL(
                "INSERT INTO {0}.{1} "
                "VALUES (%(application_name)s, %(notification_id)s) "
                "ON CONFLICT (application_name) DO UPDATE "
                "SET notification_id = %(notification_id)s "
                "WHERE {0}.{1}.notification_id < %(notification_id)s "
                "RETURNING notification_id"
            ).format(
                Identifier(self.datastore.schema),
                Identifier(self.tracking_table_name),
            )
        else:
            # For legacy multi-row tracking.
            self.create_table_statements.append(
                SQL(
                    "CREATE TABLE IF NOT EXISTS {0}.{1} ("
                    "application_name text, "
                    "notification_id bigint, "
                    "PRIMARY KEY "
                    "(application_name, notification_id))"
                ).format(
                    Identifier(self.datastore.schema),
                    Identifier(self.tracking_table_name),
                )
            )
            self.insert_tracking_statement = SQL(
                "INSERT INTO {0}.{1} VALUES (%(application_name)s, %(notification_id)s)"
            ).format(
                Identifier(self.datastore.schema),
                Identifier(self.tracking_table_name),
            )

        self.max_tracking_id_statement = SQL(
            "SELECT MAX(notification_id) FROM {0}.{1} WHERE application_name=%s"
        ).format(
            Identifier(self.datastore.schema),
            Identifier(self.tracking_table_name),
        )

    def create_table(self) -> None:
        # Get the migration version.
        try:
            self.tracking_migration_current = self.tracking_migration_previous = (
                self.max_tracking_id(self.table_migration_identifier)
            )
        except ProgrammingError:
            pass
        else:
            self.tracking_table_exists = True
        super().create_table()
        if (
            not self.datastore.single_row_tracking
            and self.tracking_migration_current is not None
        ):
            msg = "Can't do multi-row tracking with single-row tracking table"
            raise OperationalError(msg)

    def _create_table(self, curs: Cursor[DictRow]) -> None:
        max_tracking_ids: dict[str, int] = {}
        if (
            self.datastore.single_row_tracking
            and self.tracking_table_exists
            and not self.tracking_migration_previous
        ):
            # Migrate the table.
            curs.execute(
                SQL("SET LOCAL lock_timeout = '{0}s'").format(
                    self.datastore.lock_timeout
                )
            )
            curs.execute(
                SQL("LOCK TABLE {0}.{1} IN ACCESS EXCLUSIVE MODE").format(
                    Identifier(self.datastore.schema),
                    Identifier(self.tracking_table_name),
                )
            )

            # Get all application names.
            application_names: list[str] = [
                select_row["application_name"]
                for select_row in curs.execute(
                    SQL("SELECT DISTINCT application_name FROM {0}.{1}").format(
                        Identifier(self.datastore.schema),
                        Identifier(self.tracking_table_name),
                    )
                )
            ]

            # Get max tracking ID for each application name.
            for application_name in application_names:
                curs.execute(self.max_tracking_id_statement, (application_name,))
                max_tracking_id_row = curs.fetchone()
                assert max_tracking_id_row is not None
                max_tracking_ids[application_name] = max_tracking_id_row["max"]
            # Rename the table.
            rename = f"bkup1_{self.tracking_table_name}"[: self.MAX_IDENTIFIER_LEN]
            drop_table_statement = SQL("ALTER TABLE {0}.{1} RENAME TO {2}").format(
                Identifier(self.datastore.schema),
                Identifier(self.tracking_table_name),
                Identifier(rename),
            )
            curs.execute(drop_table_statement)
        # Create the table.
        super()._create_table(curs)
        # Maybe insert migration tracking record and application tracking records.
        if self.datastore.single_row_tracking and (
            not self.tracking_table_exists
            or (self.tracking_table_exists and not self.tracking_migration_previous)
        ):
            # Assume we just created a table for single-row tracking.
            self._insert_tracking(curs, Tracking(self.table_migration_identifier, 1))
            self.tracking_migration_current = 1
            for application_name, max_tracking_id in max_tracking_ids.items():
                self._insert_tracking(curs, Tracking(application_name, max_tracking_id))

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def insert_tracking(self, tracking: Tracking) -> None:
        with self.datastore.transaction(commit=True) as curs:
            self._insert_tracking(curs, tracking)

    def _insert_tracking(
        self,
        curs: Cursor[DictRow],
        tracking: Tracking,
    ) -> None:
        self._check_has_multi_row_tracking_table(curs)

        curs.execute(
            query=self.insert_tracking_statement,
            params={
                "application_name": tracking.application_name,
                "notification_id": tracking.notification_id,
            },
            prepare=True,
        )
        if self.datastore.single_row_tracking:
            fetchone = curs.fetchone()
            if fetchone is None:
                msg = (
                    "Failed to record tracking for "
                    f"{tracking.application_name} {tracking.notification_id}"
                )
                raise IntegrityError(msg)

    def _check_has_multi_row_tracking_table(self, c: Cursor[DictRow]) -> None:
        if (
            not self.datastore.single_row_tracking
            and not self.has_checked_for_multi_row_tracking_table
            and self._max_tracking_id(self.table_migration_identifier, c)
        ):
            msg = "Can't do multi-row tracking with single-row tracking table"
            raise ProgrammingError(msg)
        self.has_checked_for_multi_row_tracking_table = True

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def max_tracking_id(self, application_name: str) -> int | None:
        with self.datastore.get_connection() as conn, conn.cursor() as curs:
            return self._max_tracking_id(application_name, curs)

    def _max_tracking_id(
        self, application_name: str, curs: Cursor[DictRow]
    ) -> int | None:
        curs.execute(
            query=self.max_tracking_id_statement,
            params=(application_name,),
            prepare=True,
        )
        fetchone = curs.fetchone()
        assert fetchone is not None
        return fetchone["max"]

    @retry((InterfaceError, OperationalError), max_attempts=10, wait=0.2)
    def has_tracking_id(
        self, application_name: str, notification_id: int | None
    ) -> bool:
        return super().has_tracking_id(application_name, notification_id)


TPostgresTrackingRecorder = TypeVar(
    "TPostgresTrackingRecorder",
    bound=PostgresTrackingRecorder,
    default=PostgresTrackingRecorder,
)


class PostgresProcessRecorder(
    PostgresTrackingRecorder, PostgresApplicationRecorder, ProcessRecorder
):
    def __init__(
        self,
        datastore: PostgresDatastore,
        *,
        events_table_name: str = "stored_events",
        tracking_table_name: str = "notification_tracking",
    ):
        super().__init__(
            datastore,
            tracking_table_name=tracking_table_name,
            events_table_name=events_table_name,
        )

    def _insert_events(
        self,
        curs: Cursor[DictRow],
        stored_events: Sequence[StoredEvent],
        **kwargs: Any,
    ) -> None:
        tracking: Tracking | None = kwargs.get("tracking")
        if tracking is not None:
            self._insert_tracking(curs, tracking=tracking)
        super()._insert_events(curs, stored_events, **kwargs)


class PostgresFactory(InfrastructureFactory[PostgresTrackingRecorder]):
    POSTGRES_DBNAME = "POSTGRES_DBNAME"
    POSTGRES_HOST = "POSTGRES_HOST"
    POSTGRES_PORT = "POSTGRES_PORT"
    POSTGRES_USER = "POSTGRES_USER"
    POSTGRES_PASSWORD = "POSTGRES_PASSWORD"  # noqa: S105
    POSTGRES_GET_PASSWORD_TOPIC = "POSTGRES_GET_PASSWORD_TOPIC"  # noqa: S105
    POSTGRES_CONNECT_TIMEOUT = "POSTGRES_CONNECT_TIMEOUT"
    POSTGRES_CONN_MAX_AGE = "POSTGRES_CONN_MAX_AGE"
    POSTGRES_PRE_PING = "POSTGRES_PRE_PING"
    POSTGRES_MAX_WAITING = "POSTGRES_MAX_WAITING"
    POSTGRES_LOCK_TIMEOUT = "POSTGRES_LOCK_TIMEOUT"
    POSTGRES_POOL_SIZE = "POSTGRES_POOL_SIZE"
    POSTGRES_MAX_OVERFLOW = "POSTGRES_MAX_OVERFLOW"
    POSTGRES_IDLE_IN_TRANSACTION_SESSION_TIMEOUT = (
        "POSTGRES_IDLE_IN_TRANSACTION_SESSION_TIMEOUT"
    )
    POSTGRES_SCHEMA = "POSTGRES_SCHEMA"
    POSTGRES_SINGLE_ROW_TRACKING = "SINGLE_ROW_TRACKING"
    CREATE_TABLE = "CREATE_TABLE"

    aggregate_recorder_class = PostgresAggregateRecorder
    application_recorder_class = PostgresApplicationRecorder
    tracking_recorder_class = PostgresTrackingRecorder
    process_recorder_class = PostgresProcessRecorder

    def __init__(self, env: Environment | EnvType | None):
        super().__init__(env)
        dbname = self.env.get(self.POSTGRES_DBNAME)
        if dbname is None:
            msg = (
                "Postgres database name not found "
                "in environment with key "
                f"'{self.POSTGRES_DBNAME}'"
            )
            # TODO: Indicate both keys here, also for other environment variables.
            # ) + " or ".join(
            #   [f"'{key}'" for key in self.env.create_keys(self.POSTGRES_DBNAME)]
            # )
            raise OSError(msg)

        host = self.env.get(self.POSTGRES_HOST)
        if host is None:
            msg = (
                "Postgres host not found "
                "in environment with key "
                f"'{self.POSTGRES_HOST}'"
            )
            raise OSError(msg)

        port = self.env.get(self.POSTGRES_PORT) or "5432"

        user = self.env.get(self.POSTGRES_USER)
        if user is None:
            msg = (
                "Postgres user not found "
                "in environment with key "
                f"'{self.POSTGRES_USER}'"
            )
            raise OSError(msg)

        get_password_func = None
        get_password_topic = self.env.get(self.POSTGRES_GET_PASSWORD_TOPIC)
        if not get_password_topic:
            password = self.env.get(self.POSTGRES_PASSWORD)
            if password is None:
                msg = (
                    "Postgres password not found "
                    "in environment with key "
                    f"'{self.POSTGRES_PASSWORD}'"
                )
                raise OSError(msg)
        else:
            get_password_func = resolve_topic(get_password_topic)
            password = ""

        connect_timeout = 30
        connect_timeout_str = self.env.get(self.POSTGRES_CONNECT_TIMEOUT)
        if connect_timeout_str:
            try:
                connect_timeout = int(connect_timeout_str)
            except ValueError:
                msg = (
                    "Postgres environment value for key "
                    f"'{self.POSTGRES_CONNECT_TIMEOUT}' is invalid. "
                    "If set, an integer or empty string is expected: "
                    f"'{connect_timeout_str}'"
                )
                raise OSError(msg) from None

        idle_in_transaction_session_timeout_str = (
            self.env.get(self.POSTGRES_IDLE_IN_TRANSACTION_SESSION_TIMEOUT) or "5"
        )

        try:
            idle_in_transaction_session_timeout = int(
                idle_in_transaction_session_timeout_str
            )
        except ValueError:
            msg = (
                "Postgres environment value for key "
                f"'{self.POSTGRES_IDLE_IN_TRANSACTION_SESSION_TIMEOUT}' is invalid. "
                "If set, an integer or empty string is expected: "
                f"'{idle_in_transaction_session_timeout_str}'"
            )
            raise OSError(msg) from None

        pool_size = 5
        pool_size_str = self.env.get(self.POSTGRES_POOL_SIZE)
        if pool_size_str:
            try:
                pool_size = int(pool_size_str)
            except ValueError:
                msg = (
                    "Postgres environment value for key "
                    f"'{self.POSTGRES_POOL_SIZE}' is invalid. "
                    "If set, an integer or empty string is expected: "
                    f"'{pool_size_str}'"
                )
                raise OSError(msg) from None

        pool_max_overflow = 10
        pool_max_overflow_str = self.env.get(self.POSTGRES_MAX_OVERFLOW)
        if pool_max_overflow_str:
            try:
                pool_max_overflow = int(pool_max_overflow_str)
            except ValueError:
                msg = (
                    "Postgres environment value for key "
                    f"'{self.POSTGRES_MAX_OVERFLOW}' is invalid. "
                    "If set, an integer or empty string is expected: "
                    f"'{pool_max_overflow_str}'"
                )
                raise OSError(msg) from None

        max_waiting = 0
        max_waiting_str = self.env.get(self.POSTGRES_MAX_WAITING)
        if max_waiting_str:
            try:
                max_waiting = int(max_waiting_str)
            except ValueError:
                msg = (
                    "Postgres environment value for key "
                    f"'{self.POSTGRES_MAX_WAITING}' is invalid. "
                    "If set, an integer or empty string is expected: "
                    f"'{max_waiting_str}'"
                )
                raise OSError(msg) from None

        conn_max_age = 60 * 60.0
        conn_max_age_str = self.env.get(self.POSTGRES_CONN_MAX_AGE)
        if conn_max_age_str:
            try:
                conn_max_age = float(conn_max_age_str)
            except ValueError:
                msg = (
                    "Postgres environment value for key "
                    f"'{self.POSTGRES_CONN_MAX_AGE}' is invalid. "
                    "If set, a float or empty string is expected: "
                    f"'{conn_max_age_str}'"
                )
                raise OSError(msg) from None

        pre_ping = strtobool(self.env.get(self.POSTGRES_PRE_PING) or "no")

        lock_timeout_str = self.env.get(self.POSTGRES_LOCK_TIMEOUT) or "0"

        try:
            lock_timeout = int(lock_timeout_str)
        except ValueError:
            msg = (
                "Postgres environment value for key "
                f"'{self.POSTGRES_LOCK_TIMEOUT}' is invalid. "
                "If set, an integer or empty string is expected: "
                f"'{lock_timeout_str}'"
            )
            raise OSError(msg) from None

        schema = self.env.get(self.POSTGRES_SCHEMA) or ""

        single_row_tracking = strtobool(
            self.env.get(self.POSTGRES_SINGLE_ROW_TRACKING, "t")
        )

        self.datastore = PostgresDatastore(
            dbname=dbname,
            host=host,
            port=port,
            user=user,
            password=password,
            connect_timeout=connect_timeout,
            idle_in_transaction_session_timeout=idle_in_transaction_session_timeout,
            pool_size=pool_size,
            max_overflow=pool_max_overflow,
            max_waiting=max_waiting,
            conn_max_age=conn_max_age,
            pre_ping=pre_ping,
            lock_timeout=lock_timeout,
            schema=schema,
            get_password_func=get_password_func,
            single_row_tracking=single_row_tracking,
        )

    def env_create_table(self) -> bool:
        return strtobool(self.env.get(self.CREATE_TABLE) or "yes")

    def aggregate_recorder(self, purpose: str = "events") -> AggregateRecorder:
        prefix = self.env.name.lower() or "stored"
        events_table_name = prefix + "_" + purpose
        recorder = type(self).aggregate_recorder_class(
            datastore=self.datastore,
            events_table_name=events_table_name,
        )
        if self.env_create_table():
            recorder.create_table()
        return recorder

    def application_recorder(self) -> ApplicationRecorder:
        prefix = self.env.name.lower() or "stored"
        events_table_name = prefix + "_events"
        application_recorder_topic = self.env.get(self.APPLICATION_RECORDER_TOPIC)
        if application_recorder_topic:
            application_recorder_class: type[PostgresApplicationRecorder] = (
                resolve_topic(application_recorder_topic)
            )
            assert issubclass(application_recorder_class, PostgresApplicationRecorder)
        else:
            application_recorder_class = type(self).application_recorder_class

        recorder = application_recorder_class(
            datastore=self.datastore,
            events_table_name=events_table_name,
        )
        if self.env_create_table():
            recorder.create_table()
        return recorder

    def tracking_recorder(
        self, tracking_recorder_class: type[TPostgresTrackingRecorder] | None = None
    ) -> TPostgresTrackingRecorder:
        prefix = self.env.name.lower() or "notification"
        tracking_table_name = prefix + "_tracking"
        if tracking_recorder_class is None:
            tracking_recorder_topic = self.env.get(self.TRACKING_RECORDER_TOPIC)
            if tracking_recorder_topic:
                tracking_recorder_class = resolve_topic(tracking_recorder_topic)
            else:
                tracking_recorder_class = cast(
                    "type[TPostgresTrackingRecorder]",
                    type(self).tracking_recorder_class,
                )
        assert tracking_recorder_class is not None
        assert issubclass(tracking_recorder_class, PostgresTrackingRecorder)
        recorder = tracking_recorder_class(
            datastore=self.datastore,
            tracking_table_name=tracking_table_name,
        )
        if self.env_create_table():
            recorder.create_table()
        return recorder

    def process_recorder(self) -> ProcessRecorder:
        prefix = self.env.name.lower() or "stored"
        events_table_name = prefix + "_events"
        prefix = self.env.name.lower() or "notification"
        tracking_table_name = prefix + "_tracking"
        process_recorder_topic = self.env.get(self.PROCESS_RECORDER_TOPIC)
        if process_recorder_topic:
            process_recorder_class: type[PostgresTrackingRecorder] = resolve_topic(
                process_recorder_topic
            )
            assert issubclass(process_recorder_class, PostgresProcessRecorder)
        else:
            process_recorder_class = type(self).process_recorder_class

        recorder = process_recorder_class(
            datastore=self.datastore,
            events_table_name=events_table_name,
            tracking_table_name=tracking_table_name,
        )
        if self.env_create_table():
            recorder.create_table()
        return recorder

    def close(self) -> None:
        with contextlib.suppress(AttributeError):
            self.datastore.close()


Factory = PostgresFactory



import os

import psycopg
from psycopg.sql import SQL, Identifier

from eventsourcing.postgres import PostgresDatastore


def pg_close_all_connections(
    name: str = "eventsourcing",
    host: str = "127.0.0.1",
    port: str = "5432",
    user: str = "postgres",
    password: str = "postgres",  # noqa: S107
) -> None:
    try:
        # For local development... probably.
        pg_conn = psycopg.connect(
            dbname=name,
            host=host,
            port=port,
        )
    except psycopg.Error:
        # For GitHub actions.
        """CREATE ROLE postgres LOGIN SUPERUSER PASSWORD 'postgres';"""
        pg_conn = psycopg.connect(
            dbname=name,
            host=host,
            port=port,
            user=user,
            password=password,
        )
    close_all_connections = """
    SELECT
        pg_terminate_backend(pid)
    FROM
        pg_stat_activity
    WHERE
        -- don't kill my own connection!
        pid <> pg_backend_pid();

    """
    pg_conn_cursor = pg_conn.cursor()
    pg_conn_cursor.execute(close_all_connections)


def drop_tables() -> None:

    for schema in ["public", "myschema"]:
        datastore = PostgresDatastore(
            dbname=os.environ.get("POSTGRES_DBNAME", "eventsourcing"),
            host=os.environ.get("POSTGRES_HOST", "127.0.0.1"),
            port=os.environ.get("POSTGRES_PORT", "5432"),
            user=os.environ.get("POSTGRES_USER", "eventsourcing"),
            password=os.environ.get("POSTGRES_PASSWORD", "eventsourcing"),
            schema=schema,
        )
        with datastore.transaction(commit=True) as curs:
            select_table_names = SQL(
                "SELECT table_name FROM information_schema.tables "
                "WHERE table_schema = %s"
            )
            fetchall = curs.execute(select_table_names, (datastore.schema,)).fetchall()
            for row in fetchall:
                table_name = row["table_name"]
                # print(f"Dropping table '{table_name}' in schema '{schema}'")
                statement = SQL("DROP TABLE IF EXISTS {0}.{1}").format(
                    Identifier(datastore.schema), Identifier(table_name)
                )
                curs.execute(statement, prepare=False)
                # print(f"Dropped table '{table_name}' in schema '{schema}'")



from __future__ import annotations

import traceback
import warnings
from concurrent.futures import ThreadPoolExecutor
from decimal import Decimal
from threading import Event, get_ident
from time import sleep
from typing import TYPE_CHECKING, Any, ClassVar
from unittest import TestCase
from uuid import UUID, uuid4

from eventsourcing.application import AggregateNotFoundError, Application
from eventsourcing.domain import Aggregate
from eventsourcing.persistence import (
    InfrastructureFactory,
    InfrastructureFactoryError,
    IntegrityError,
    JSONTranscoder,
    Transcoding,
)
from eventsourcing.tests.domain import BankAccount, EmailAddress
from eventsourcing.utils import EnvType, get_topic

if TYPE_CHECKING:
    from datetime import datetime


class ExampleApplicationTestCase(TestCase):
    started_ats: ClassVar[dict[type[TestCase], datetime]] = {}
    counts: ClassVar[dict[type[TestCase], int]] = {}
    expected_factory_topic: str

    def test_example_application(self) -> None:
        app = BankAccounts(env={"IS_SNAPSHOTTING_ENABLED": "y"})

        self.assertEqual(get_topic(type(app.factory)), self.expected_factory_topic)

        # Check AccountNotFound exception.
        with self.assertRaises(BankAccounts.AccountNotFoundError):
            app.get_account(uuid4())

        # Open an account.
        account_id = app.open_account(
            full_name="Alice",
            email_address="alice@example.com",
        )

        # Check balance.
        self.assertEqual(
            app.get_balance(account_id),
            Decimal("0.00"),
        )

        # Credit the account.
        app.credit_account(account_id, Decimal("10.00"))

        # Check balance.
        self.assertEqual(
            app.get_balance(account_id),
            Decimal("10.00"),
        )

        app.credit_account(account_id, Decimal("25.00"))
        app.credit_account(account_id, Decimal("30.00"))

        # Check balance.
        self.assertEqual(
            app.get_balance(account_id),
            Decimal("65.00"),
        )

        # sleep(1)  # Added to make eventsourcing-axon tests work.
        section = app.notification_log["1,10"]
        self.assertEqual(len(section.items), 4)

        # Take snapshot (specify version).
        app.take_snapshot(account_id, version=Aggregate.INITIAL_VERSION + 1)

        assert app.snapshots is not None  # for mypy
        snapshots = list(app.snapshots.get(account_id))
        self.assertEqual(len(snapshots), 1)
        self.assertEqual(snapshots[0].originator_version, Aggregate.INITIAL_VERSION + 1)

        from_snapshot1: BankAccount = app.repository.get(
            account_id, version=Aggregate.INITIAL_VERSION + 2
        )
        self.assertIsInstance(from_snapshot1, BankAccount)
        self.assertEqual(from_snapshot1.version, Aggregate.INITIAL_VERSION + 2)
        self.assertEqual(from_snapshot1.balance, Decimal("35.00"))

        # Take snapshot (don't specify version).
        app.take_snapshot(account_id)
        assert app.snapshots is not None  # for mypy
        snapshots = list(app.snapshots.get(account_id))
        self.assertEqual(len(snapshots), 2)
        self.assertEqual(snapshots[0].originator_version, Aggregate.INITIAL_VERSION + 1)
        self.assertEqual(snapshots[1].originator_version, Aggregate.INITIAL_VERSION + 3)

        from_snapshot2: BankAccount = app.repository.get(account_id)
        self.assertIsInstance(from_snapshot2, BankAccount)
        self.assertEqual(from_snapshot2.version, Aggregate.INITIAL_VERSION + 3)
        self.assertEqual(from_snapshot2.balance, Decimal("65.00"))


class EmailAddressAsStr(Transcoding):
    type = EmailAddress
    name = "email_address_as_str"

    def encode(self, obj: EmailAddress) -> str:
        return obj.address

    def decode(self, data: str) -> EmailAddress:
        return EmailAddress(data)


class BankAccounts(Application[UUID]):
    is_snapshotting_enabled = True

    def register_transcodings(self, transcoder: JSONTranscoder) -> None:
        super().register_transcodings(transcoder)
        transcoder.register(EmailAddressAsStr())

    def open_account(self, full_name: str, email_address: str) -> UUID:
        account = BankAccount.open(
            full_name=full_name,
            email_address=email_address,
        )
        self.save(account)
        return account.id

    def credit_account(self, account_id: UUID, amount: Decimal) -> None:
        account = self.get_account(account_id)
        account.append_transaction(amount)
        self.save(account)

    def get_balance(self, account_id: UUID) -> Decimal:
        account = self.get_account(account_id)
        return account.balance

    def get_account(self, account_id: UUID) -> BankAccount:
        try:
            aggregate: BankAccount = self.repository.get(account_id)
        except AggregateNotFoundError:
            raise self.AccountNotFoundError(account_id) from None
        else:
            assert isinstance(aggregate, BankAccount)
            return aggregate

    class AccountNotFoundError(Exception):
        pass


class ApplicationTestCase(TestCase):
    def test_name(self) -> None:
        self.assertEqual(Application.name, "Application")

        class MyApplication1(Application[UUID]):
            pass

        self.assertEqual(MyApplication1.name, "MyApplication1")

        class MyApplication2(Application[UUID]):
            name = "MyBoundedContext"

        self.assertEqual(MyApplication2.name, "MyBoundedContext")

    def test_resolve_persistence_topics(self) -> None:
        # None specified.
        app = Application[UUID]()
        self.assertIsInstance(app.factory, InfrastructureFactory)

        # Legacy 'INFRASTRUCTURE_FACTORY'.
        app = Application(env={"INFRASTRUCTURE_FACTORY": "eventsourcing.popo:Factory"})
        self.assertIsInstance(app.factory, InfrastructureFactory)

        # Legacy 'FACTORY_TOPIC'.
        app = Application(env={"FACTORY_TOPIC": "eventsourcing.popo:Factory"})
        self.assertIsInstance(app.factory, InfrastructureFactory)

        # Check 'PERSISTENCE_MODULE' resolves to a class.
        app = Application(env={"PERSISTENCE_MODULE": "eventsourcing.popo"})
        self.assertIsInstance(app.factory, InfrastructureFactory)

        # Check exceptions.
        with self.assertRaises(InfrastructureFactoryError) as cm:
            Application(env={"PERSISTENCE_MODULE": "eventsourcing.application"})
        self.assertEqual(
            cm.exception.args[0],
            "Found 0 infrastructure factory classes in "
            "'eventsourcing.application', expected 1.",
        )

        with self.assertRaises(InfrastructureFactoryError) as cm:
            Application(
                env={"PERSISTENCE_MODULE": "eventsourcing.application:Application"}
            )
        self.assertEqual(
            "Topic 'eventsourcing.application:Application' didn't "
            "resolve to a persistence module or infrastructure factory class: "
            "<class 'eventsourcing.application.Application'>",
            cm.exception.args[0],
        )

    def test_save_returns_recording_event(self) -> None:
        app = Application[UUID]()

        recordings = app.save()
        self.assertEqual(recordings, [])

        recordings = app.save(None)
        self.assertEqual(recordings, [])

        recordings = app.save(Aggregate())
        self.assertEqual(len(recordings), 1)
        self.assertEqual(recordings[0].notification.id, 1)

        recordings = app.save(Aggregate())
        self.assertEqual(len(recordings), 1)
        self.assertEqual(recordings[0].notification.id, 2)

        recordings = app.save(Aggregate(), Aggregate())
        self.assertEqual(len(recordings), 2)
        self.assertEqual(recordings[0].notification.id, 3)
        self.assertEqual(recordings[1].notification.id, 4)

    def test_take_snapshot_raises_assertion_error_if_snapshotting_not_enabled(
        self,
    ) -> None:
        app = Application[UUID]()
        with self.assertRaises(AssertionError) as cm:
            app.take_snapshot(uuid4())
        self.assertEqual(
            cm.exception.args[0],
            "Can't take snapshot without snapshots store. Please "
            "set environment variable IS_SNAPSHOTTING_ENABLED to "
            "a true value (e.g. 'y'), or set 'is_snapshotting_enabled' "
            "on application class, or set 'snapshotting_intervals' on "
            "application class.",
        )

    def test_application_with_cached_aggregates_and_fastforward(self) -> None:
        app = Application[UUID](env={"AGGREGATE_CACHE_MAXSIZE": "10"})

        aggregate = Aggregate()
        app.save(aggregate)
        # Should not put the aggregate in the cache.
        assert app.repository.cache is not None  # for mypy
        with self.assertRaises(KeyError):
            self.assertEqual(aggregate, app.repository.cache.get(aggregate.id))

        # Getting the aggregate should put aggregate in the cache.
        app.repository.get(aggregate.id)
        self.assertEqual(aggregate, app.repository.cache.get(aggregate.id))

        # Triggering a subsequent event shouldn't update the cache.
        aggregate.trigger_event(Aggregate.Event)
        app.save(aggregate)
        self.assertNotEqual(aggregate, app.repository.cache.get(aggregate.id))
        self.assertEqual(
            aggregate.version, app.repository.cache.get(aggregate.id).version + 1
        )

        # Getting the aggregate should fastforward the aggregate in the cache.
        app.repository.get(aggregate.id)
        self.assertEqual(aggregate, app.repository.cache.get(aggregate.id))

    def test_check_aggregate_fastforwarding_nonblocking(self) -> None:
        self._check_aggregate_fastforwarding_during_contention(
            env={
                "AGGREGATE_CACHE_MAXSIZE": "10",
                "AGGREGATE_CACHE_FASTFORWARD_SKIPPING": "y",
            }
        )

    def test_check_aggregate_fastforwarding_blocking(self) -> None:
        self._check_aggregate_fastforwarding_during_contention(
            env={"AGGREGATE_CACHE_MAXSIZE": "10"}
        )

    def _check_aggregate_fastforwarding_during_contention(self, env: EnvType) -> None:
        app = Application[UUID](env=env)

        self.assertEqual(len(app.repository._fastforward_locks_inuse), 0)

        # Create one aggregate.
        original_aggregate = Aggregate()
        app.save(original_aggregate)
        obj_ids = set()

        # Prime the cache.
        app.repository.get(original_aggregate.id)

        # Remember the aggregate ID.
        aggregate_id = original_aggregate.id

        stopped = Event()
        errors: list[BaseException] = []
        successful_thread_ids = set()

        def trigger_save_get_check() -> None:
            while not stopped.is_set():
                try:
                    # Get the aggregate.
                    aggregate: Aggregate = app.repository.get(aggregate_id)
                    original_version = aggregate.version

                    # Try to record a new event.
                    aggregate.trigger_event(Aggregate.Event)
                    # Give other threads a chance.
                    try:
                        app.save(aggregate)
                    except IntegrityError:
                        # Start again if we didn't record a new event.
                        # print("Got integrity error")
                        sleep(0.001)
                        continue

                    # Get the aggregate from the cache.
                    assert app.repository.cache is not None
                    cached: Any = app.repository.cache.get(aggregate_id)
                    obj_ids.add(id(cached))

                    if len(obj_ids) > 1:
                        stopped.set()
                        continue

                    # Fast-forward the cached aggregate.
                    fastforwarded: Aggregate = app.repository.get(aggregate_id)

                    # Check cached aggregate was fast-forwarded with recorded event.
                    if fastforwarded.version < original_version:
                        try:
                            self.fail(
                                f"Failed to fast-forward at version {original_version}"
                            )
                        except AssertionError as e:
                            errors.append(e)
                            stopped.set()
                            continue

                    # Monitor number of threads getting involved.
                    thread_id = get_ident()
                    successful_thread_ids.add(thread_id)

                    # print("Version:", aggregate.version, thread_id)

                    # See if we have done enough.
                    if len(successful_thread_ids) > 10 and aggregate.version >= 25:
                        stopped.set()
                        continue

                    sleep(0.0001)
                    # sleep(0.001)
                except BaseException as e:
                    errors.append(e)
                    stopped.set()
                    print(traceback.format_exc())
                    raise

        executor = ThreadPoolExecutor(max_workers=100)
        futures = []
        for _ in range(100):
            f = executor.submit(trigger_save_get_check)
            futures.append(f)

        # Run for three seconds.
        stopped.wait(timeout=10)
        for f in futures:
            f.result()
        # print("Got all results, shutting down executor")
        executor.shutdown()

        try:
            if errors:
                raise errors[0]
            if len(obj_ids) > 1:
                self.fail(f"More than one instance used in the cache: {len(obj_ids)}")
            if len(successful_thread_ids) < 3:
                self.fail("Insufficient sharing across contentious threads")

            final_aggregate: Aggregate = app.repository.get(aggregate_id)
            # print("Final aggregate version:", final_aggregate.version)
            if final_aggregate.version < 25:
                self.fail(f"Insufficient version increment: {final_aggregate.version}")

            self.assertEqual(len(app.repository._fastforward_locks_inuse), 0)

        finally:
            # print("Closing application")
            app.close()

    def test_application_with_cached_aggregates_not_fastforward(self) -> None:
        app = Application[UUID](
            env={
                "AGGREGATE_CACHE_MAXSIZE": "10",
                "AGGREGATE_CACHE_FASTFORWARD": "f",
            }
        )
        aggregate1 = Aggregate()
        app.save(aggregate1)
        aggregate_id = aggregate1.id

        # Should put the aggregate in the cache.
        assert app.repository.cache is not None  # for mypy
        self.assertEqual(aggregate1, app.repository.cache.get(aggregate_id))
        app.repository.get(aggregate_id)
        self.assertEqual(aggregate1, app.repository.cache.get(aggregate_id))

        aggregate2 = Aggregate()
        aggregate2._id = aggregate_id
        aggregate2.trigger_event(Aggregate.Event)

        # This will replace object in cache.
        app.save(aggregate2)

        self.assertEqual(aggregate2.version, aggregate1.version + 1)
        aggregate3: Aggregate = app.repository.get(aggregate_id)
        self.assertEqual(aggregate3.version, aggregate3.version)
        self.assertEqual(id(aggregate3.version), id(aggregate3.version))

        # This will mess things up because the cache has a stale aggregate.
        aggregate3.trigger_event(Aggregate.Event)
        app.events.put(aggregate3.collect_events())

        # And so using the aggregate to record new events will cause an IntegrityError.
        aggregate4: Aggregate = app.repository.get(aggregate_id)
        aggregate4.trigger_event(Aggregate.Event)
        with self.assertRaises(IntegrityError):
            app.save(aggregate4)

    def test_application_with_deepcopy_from_cache_arg(self) -> None:
        app = Application[UUID](
            env={
                "AGGREGATE_CACHE_MAXSIZE": "10",
            }
        )
        aggregate = Aggregate()
        app.save(aggregate)
        self.assertEqual(aggregate.version, 1)
        reconstructed: Aggregate = app.repository.get(aggregate.id)
        reconstructed.version = 101
        assert app.repository.cache is not None  # for mypy
        self.assertEqual(app.repository.cache.get(aggregate.id).version, 1)
        cached: Aggregate = app.repository.get(aggregate.id, deepcopy_from_cache=False)
        cached.version = 101
        self.assertEqual(app.repository.cache.get(aggregate.id).version, 101)

    def test_application_with_deepcopy_from_cache_attribute(self) -> None:
        app = Application[UUID](
            env={
                "AGGREGATE_CACHE_MAXSIZE": "10",
            }
        )
        aggregate = Aggregate()
        app.save(aggregate)
        self.assertEqual(aggregate.version, 1)
        reconstructed: Aggregate = app.repository.get(aggregate.id)
        reconstructed.version = 101
        assert app.repository.cache is not None  # for mypy
        self.assertEqual(app.repository.cache.get(aggregate.id).version, 1)
        app.repository.deepcopy_from_cache = False
        cached: Aggregate = app.repository.get(aggregate.id)
        cached.version = 101
        self.assertEqual(app.repository.cache.get(aggregate.id).version, 101)

    def test_application_log(self) -> None:
        # Check the old 'log' attribute presents the 'notification log' object.
        app = Application[UUID]()

        # Verify deprecation warning.
        with warnings.catch_warnings(record=True) as w:
            self.assertIs(app.log, app.notification_log)

        self.assertEqual(1, len(w))
        self.assertIs(w[-1].category, DeprecationWarning)
        self.assertIn(
            "'log' is deprecated, use 'notifications' instead", str(w[-1].message)
        )



from __future__ import annotations

import contextlib
import os
import threading
import weakref
from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from threading import Event, Thread
from traceback import format_exc
from typing import TYPE_CHECKING, Any, Generic, TypeVar
from warnings import warn

from eventsourcing.application import Application, ProcessingEvent
from eventsourcing.dispatch import singledispatchmethod
from eventsourcing.domain import DomainEventProtocol, TAggregateID
from eventsourcing.persistence import (
    InfrastructureFactory,
    Mapper,
    ProcessRecorder,
    Tracking,
    TrackingRecorder,
    TTrackingRecorder,
    WaitInterruptedError,
)
from eventsourcing.utils import Environment, EnvType

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self


class ApplicationSubscription(
    Iterator[tuple[DomainEventProtocol[TAggregateID], Tracking]]
):
    """An iterator that yields all domain events recorded in an application
    sequence that have notification IDs greater than a given value. The iterator
    will block when all recorded domain events have been yielded, and then
    continue when new events are recorded. Domain events are returned along
    with tracking objects that identify the position in the application sequence.
    """

    def __init__(
        self,
        app: Application[TAggregateID],
        gt: int | None = None,
        topics: Sequence[str] = (),
    ):
        """
        Starts a subscription to application's recorder.
        """
        self.name = app.name
        self.recorder = app.recorder
        self.mapper: Mapper[TAggregateID] = app.mapper
        self.subscription = self.recorder.subscribe(gt=gt, topics=topics)

    def stop(self) -> None:
        """Stops the subscription to the application's recorder."""
        self.subscription.stop()

    def __enter__(self) -> Self:
        """Calls __enter__ on the stored event subscription."""
        self.subscription.__enter__()
        return self

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        """Calls __exit__ on the stored event subscription."""
        self.subscription.__exit__(*args, **kwargs)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> tuple[DomainEventProtocol[TAggregateID], Tracking]:
        """Returns the next stored event from subscription to the application's
        recorder. Constructs a tracking object that identifies the position of
        the event in the application sequence. Constructs a domain event object
        from the stored event object using the application's mapper. Returns a
        tuple of the domain event object and the tracking object.
        """
        notification = next(self.subscription)
        tracking = Tracking(self.name, notification.id)
        domain_event = self.mapper.to_domain_event(notification)
        return domain_event, tracking

    def __del__(self) -> None:
        """Stops the stored event subscription."""
        with contextlib.suppress(AttributeError):
            self.stop()


class Projection(ABC, Generic[TTrackingRecorder]):
    name: str = ""
    """
    Name of projection, used to pick prefixed environment
    variables and define database table names.
    """
    topics: tuple[str, ...] = ()
    """
    Event topics, used to filter events in database when subscribing to an application.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if "name" not in cls.__dict__:
            cls.name = cls.__name__

    def __init__(
        self,
        view: TTrackingRecorder,
    ):
        """Initialises the view property with the given view argument."""
        self._view = view

    @property
    def view(self) -> TTrackingRecorder:
        """Materialised view of an event-sourced application."""
        return self._view

    @singledispatchmethod
    @abstractmethod
    def process_event(
        self, domain_event: DomainEventProtocol[TAggregateID], tracking: Tracking
    ) -> None:
        """Process a domain event and track it."""


class EventSourcedProjection(Application[TAggregateID], ABC):
    """Extends the :py:class:`~eventsourcing.application.Application` class
    by using a process recorder as its application recorder, and by
    processing domain events through its :py:func:`policy` method.
    """

    topics: Sequence[str] = ()

    def __init__(self, env: EnvType | None = None) -> None:
        super().__init__(env)
        self.recorder: ProcessRecorder
        self.processing_lock = threading.Lock()

    def construct_recorder(self) -> ProcessRecorder:
        """Constructs and returns a :class:`~eventsourcing.persistence.ProcessRecorder`
        for the application to use as its application recorder.
        """
        return self.factory.process_recorder()

    def process_event(
        self, domain_event: DomainEventProtocol[TAggregateID], tracking: Tracking
    ) -> None:
        """Calls :func:`~eventsourcing.system.Follower.policy` method with the given
        domain event and a new :class:`~eventsourcing.application.ProcessingEvent`
        constructed with the given tracking object.

        The policy method should collect any new aggregate events on the process
        event object.

        After the policy method returns, the processing event object will be recorded
        by calling :py:func:`~eventsourcing.application.Application._record`,
        which then returns list of :py:class:`~eventsourcing.persistence.Recording`.

        After calling :func:`~eventsourcing.application.Application._take_snapshots`,
        the recordings are passed in a call to
        :py:func:`~eventsourcing.application.Application._notify`.
        """
        processing_event = ProcessingEvent[TAggregateID](tracking=tracking)
        self.policy(domain_event, processing_event)
        recordings = self._record(processing_event)
        self._take_snapshots(processing_event)
        self.notify(processing_event.events)
        self._notify(recordings)

    @singledispatchmethod
    def policy(
        self,
        domain_event: DomainEventProtocol[TAggregateID],
        processing_event: ProcessingEvent[TAggregateID],
    ) -> None:
        """Abstract domain event processing policy method. Must be
        implemented by event processing applications. When
        processing the given domain event, event processing
        applications must use the :func:`~ProcessingEvent.collect_events`
        method of the given :py:class:`~ProcessingEvent` object (not
        the application's :func:`~eventsourcing.application.Application.save`
        method) so that the new domain events will be recorded atomically
        and uniquely with tracking information about the position of the processed
        event in its application sequence.
        """


TApplication = TypeVar("TApplication", bound=Application[Any])
TEventSourcedProjection = TypeVar(
    "TEventSourcedProjection", bound=EventSourcedProjection[Any]
)


class BaseProjectionRunner(Generic[TApplication]):
    def __init__(
        self,
        *,
        projection: EventSourcedProjection[Any] | Projection[Any],
        application_class: type[TApplication],
        tracking_recorder: TrackingRecorder,
        topics: Sequence[str],
        env: EnvType | None = None,
    ) -> None:
        self._projection = projection
        self._is_interrupted = Event()
        self._has_called_stop = False

        # Construct the application.
        self.app: TApplication = application_class(env)

        self._tracking_recorder = tracking_recorder

        # Subscribe to the application.
        self._subscription = ApplicationSubscription(
            app=self.app,
            gt=self._tracking_recorder.max_tracking_id(self.app.name),
            topics=topics,
        )

        # Start a thread to stop the subscription when the runner is interrupted.
        self._thread_error: BaseException | None = None
        self._stop_thread = Thread(
            target=self._stop_subscription_when_stopping,
            kwargs={
                "subscription": self._subscription,
                "is_stopping": self._is_interrupted,
            },
        )
        self._stop_thread.start()

        # Start a thread to iterate over the subscription.
        self._processing_thread = Thread(
            target=self._process_events_loop,
            kwargs={
                "subscription": self._subscription,
                "projection": self._projection,
                "is_stopping": self._is_interrupted,
                "runner": weakref.ref(self),
            },
        )
        self._processing_thread.start()

    @property
    def is_interrupted(self) -> Event:
        return self._is_interrupted

    @staticmethod
    def _construct_env(name: str, env: EnvType | None = None) -> Environment:
        """Constructs environment from which projection will be configured."""
        _env: dict[str, str] = {}
        _env.update(os.environ)
        if env is not None:
            _env.update(env)
        return Environment(name, _env)

    def stop(self) -> None:
        """Sets the "interrupted" event."""
        self._has_called_stop = True
        self._is_interrupted.set()

    @staticmethod
    def _stop_subscription_when_stopping(
        subscription: ApplicationSubscription[TAggregateID],
        is_stopping: Event,
    ) -> None:
        """Stops the application subscription, which
        will stop the event-processing thread.
        """
        try:
            is_stopping.wait()
        finally:
            is_stopping.set()
            subscription.stop()

    @staticmethod
    def _process_events_loop(
        subscription: ApplicationSubscription[TAggregateID],
        projection: EventSourcedProjection[Any] | Projection[Any],
        is_stopping: Event,
        runner: weakref.ReferenceType[
            ProjectionRunner[Application[Any], TrackingRecorder]
        ],
    ) -> None:
        """Iterates over the subscription and calls process_event()."""
        try:
            with subscription:
                for domain_event, tracking in subscription:
                    projection.process_event(domain_event, tracking)
        except BaseException as e:
            _runner = runner()  # get reference from weakref
            if _runner is not None:
                _runner._thread_error = e  # noqa: SLF001
            else:
                msg = "ProjectionRunner was deleted before error could be assigned:\n"
                msg += format_exc()
                warn(
                    msg,
                    RuntimeWarning,
                    stacklevel=2,
                )
        finally:
            is_stopping.set()

    def run_forever(self, timeout: float | None = None) -> None:
        """Blocks until timeout, or until the runner is stopped or errors. Re-raises
        any error otherwise exits normally
        """
        if (
            self._is_interrupted.wait(timeout=timeout)
            and self._thread_error is not None
        ):
            error = self._thread_error
            self._thread_error = None
            raise error

    def wait(self, notification_id: int | None, timeout: float = 1.0) -> None:
        """Blocks until timeout, or until the materialised view has recorded a tracking
        object that is greater than or equal to the given notification ID.
        """
        try:
            self._tracking_recorder.wait(
                application_name=self.app.name,
                notification_id=notification_id,
                timeout=timeout,
                interrupt=self._is_interrupted,
            )
        except WaitInterruptedError:
            if self._thread_error:
                error = self._thread_error
                self._thread_error = None
                raise error from None
            if self._has_called_stop:
                return
            raise

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Calls stop() and waits for the event-processing thread to exit."""
        self.stop()
        self._stop_thread.join()
        self._processing_thread.join()
        if self._thread_error:
            error = self._thread_error
            self._thread_error = None
            raise error

    def __del__(self) -> None:
        """Calls stop()."""
        with contextlib.suppress(AttributeError):
            self.stop()


class ProjectionRunner(
    BaseProjectionRunner[TApplication], Generic[TApplication, TTrackingRecorder]
):
    def __init__(
        self,
        *,
        application_class: type[TApplication],
        projection_class: type[Projection[TTrackingRecorder]],
        view_class: type[TTrackingRecorder],
        env: EnvType | None = None,
    ):
        """Constructs application from given application class with given environment.
        Also constructs a materialised view from given class using an infrastructure
        factory constructed with an environment named after the projection. Also
        constructs a projection with the constructed materialised view object.
        Starts a subscription to application and, in a separate event-processing
        thread, calls projection's process_event() method for each event and tracking
        object pair received from the subscription.
        """
        # Construct the materialised view using an infrastructure factory.
        self.view = (
            InfrastructureFactory[TTrackingRecorder]
            .construct(env=self._construct_env(name=projection_class.name, env=env))
            .tracking_recorder(view_class)
        )

        # Construct the projection using the materialised view.
        self.projection = projection_class(view=self.view)

        super().__init__(
            projection=self.projection,
            application_class=application_class,
            tracking_recorder=self.view,
            topics=self.projection.topics,
            env=env,
        )


class EventSourcedProjectionRunner(
    BaseProjectionRunner[TApplication], Generic[TApplication, TEventSourcedProjection]
):
    def __init__(
        self,
        *,
        application_class: type[TApplication],
        projection_class: type[TEventSourcedProjection],
        env: EnvType | None = None,
    ):
        self.projection: TEventSourcedProjection = projection_class(
            env=self._construct_env(name=projection_class.name, env=env)
        )

        super().__init__(
            projection=self.projection,
            application_class=application_class,
            tracking_recorder=self.projection.recorder,
            topics=self.projection.topics,
            env=env,
        )
